from __future__ import annotations

import zipfile
from pathlib import Path
from typing import Callable, Optional

import pytest
from click.testing import CliRunner, Result

from chik.cmds.beta_funcs import default_beta_root_path
from chik.cmds.chik import cli
from chik.util.beta_metrics import metrics_log_interval_default, metrics_log_interval_max, metrics_log_interval_min
from chik.util.config import lock_and_load_config, save_config


def configure(root_path: Path, *args: str) -> Result:
    return CliRunner().invoke(
        cli,
        [
            "--root-path",
            str(root_path),
            "beta",
            "configure",
            *args,
        ],
    )


def configure_interactive(root_path: Path, user_input: Optional[str] = None) -> Result:
    return CliRunner().invoke(
        cli,
        [
            "--root-path",
            str(root_path),
            "beta",
            "configure",
        ],
        input=user_input,
    )


def enable(root_path: Path, *args: str) -> Result:
    return CliRunner().invoke(
        cli,
        [
            "--root-path",
            str(root_path),
            "beta",
            "enable",
            "--force",
            *args,
        ],
    )


def enable_interactive(root_path: Path, user_input: Optional[str] = None) -> Result:
    return CliRunner().invoke(
        cli,
        [
            "--root-path",
            str(root_path),
            "beta",
            "enable",
        ],
        input=user_input,
    )


def prepare_submission(root_path: Path, user_input: Optional[str] = None) -> Result:
    return CliRunner().invoke(
        cli,
        [
            "--root-path",
            str(root_path),
            "beta",
            "prepare_submission",
        ],
        input=user_input,
    )


def generate_example_submission_data(beta_root_path: Path, versions: int, logs: int) -> None:
    for version in range(versions):
        version_path = beta_root_path / str(version)
        version_path.mkdir()
        chik_blockchain_logs = version_path / "chik-blockchain"
        plotting_logs = version_path / "plotting"
        chik_blockchain_logs.mkdir()
        plotting_logs.mkdir()
        for i in range(logs):
            with open(chik_blockchain_logs / f"beta_{i}.log", "w"):
                pass
            with open(chik_blockchain_logs / f"beta_{i + 10}.gz", "w"):
                pass
            with open(plotting_logs / f"plot_{i}.log", "w"):
                pass


def generate_beta_config(
    root_path: Path, enabled: bool, beta_path: Path, interval: int = metrics_log_interval_default
) -> None:
    with lock_and_load_config(root_path, "config.yaml") as config:
        config["beta"] = {
            "enabled": enabled,
            "path": str(beta_path),
            "metrics_log_interval": interval,
        }
        save_config(root_path, "config.yaml", config)


@pytest.mark.parametrize("interval_option", ["--interval", "-i"])
@pytest.mark.parametrize("path_option", ["--path", "-p"])
def test_configure(root_path_populated_with_config: Path, path_option: str, interval_option: str) -> None:
    root_path = root_path_populated_with_config
    beta_path = root_path / "beta"
    beta_path.mkdir()
    generate_beta_config(root_path, True, beta_path)

    result = configure(root_path, path_option, str(beta_path), interval_option, str(metrics_log_interval_max))
    assert result.exit_code == 0

    with lock_and_load_config(root_path, "config.yaml") as config:
        assert config["beta"] == {
            "enabled": True,
            "path": str(beta_path),
            "metrics_log_interval": metrics_log_interval_max,
        }


def test_configure_no_beta_config(root_path_populated_with_config: Path) -> None:
    root_path = root_path_populated_with_config
    beta_path = root_path / "beta"
    beta_path.mkdir()
    with lock_and_load_config(root_path, "config.yaml") as config:
        assert "beta" not in config

    result = configure(root_path, "--path", str(beta_path))
    assert result.exit_code == 1
    assert "beta test mode is not enabled, enable it first with `chik beta enable`" in result.output


@pytest.mark.parametrize("accept_existing_interval", [True, False])
@pytest.mark.parametrize("accept_existing_path", [True, False])
def test_beta_configure_interactive(
    root_path_populated_with_config: Path, accept_existing_path: bool, accept_existing_interval: bool
) -> None:
    assert metrics_log_interval_default != metrics_log_interval_min
    root_path = root_path_populated_with_config
    beta_path = root_path / "beta"
    generate_beta_config(root_path, True, root_path_populated_with_config)
    path_input = "\n" if accept_existing_path else str(beta_path) + "\ny\n"
    interval_input = "\n" if accept_existing_interval else str(metrics_log_interval_min) + "\n"
    result = configure_interactive(root_path, f"{path_input}{interval_input}")
    assert result.exit_code == 0
    assert "beta config updated" in result.output

    metrics_log_interval = metrics_log_interval_default if accept_existing_interval else metrics_log_interval_min
    with lock_and_load_config(root_path, "config.yaml") as config:
        assert config["beta"] == {
            "enabled": True,
            "path": str(root_path_populated_with_config if accept_existing_path else beta_path),
            "metrics_log_interval": metrics_log_interval,
        }


@pytest.mark.parametrize("option", ["--path", "-p"])
def test_beta_enable(root_path_populated_with_config: Path, option: str) -> None:
    root_path = root_path_populated_with_config
    beta_path = root_path / "beta"
    beta_path.mkdir()
    with lock_and_load_config(root_path, "config.yaml") as config:
        assert "beta" not in config

    result = enable(root_path, option, str(beta_path))
    assert result.exit_code == 0
    assert f"beta test mode enabled with path {str(beta_path)!r}" in result.output

    with lock_and_load_config(root_path, "config.yaml") as config:
        assert config["beta"] == {
            "enabled": True,
            "path": str(beta_path),
            "metrics_log_interval": metrics_log_interval_default,
        }


@pytest.mark.parametrize("enabled", [True, False])
def test_beta_enable_preconfigured(root_path_populated_with_config: Path, enabled: bool) -> None:
    root_path = root_path_populated_with_config
    beta_path = root_path / "beta"
    beta_path.mkdir()
    generate_beta_config(root_path, enabled, beta_path)

    result = enable_interactive(root_path, "y\n")

    if enabled:
        assert result.exit_code == 1
        assert "beta test mode is already enabled" in result.output
    else:
        assert result.exit_code == 0
        assert f"beta test mode enabled with path {str(beta_path)!r}" in result.output

    with lock_and_load_config(root_path, "config.yaml") as config:
        assert config["beta"] == {
            "enabled": True,
            "path": str(beta_path),
            "metrics_log_interval": metrics_log_interval_default,
        }


@pytest.mark.parametrize("accept_default_path", [True, False])
def test_beta_enable_interactive(root_path_populated_with_config: Path, accept_default_path: bool) -> None:
    root_path = root_path_populated_with_config
    beta_path = root_path / "beta"
    with lock_and_load_config(root_path, "config.yaml") as config:
        assert "beta" not in config

    result = enable_interactive(root_path, f"y\n{'' if accept_default_path else str(beta_path)}\ny\n")
    assert result.exit_code == 0
    assert (
        f"beta test mode enabled with path {str(default_beta_root_path() if accept_default_path else beta_path)!r}"
        in result.output
    )

    with lock_and_load_config(root_path, "config.yaml") as config:
        assert config["beta"] == {
            "enabled": True,
            "path": str(default_beta_root_path() if accept_default_path else beta_path),
            "metrics_log_interval": metrics_log_interval_default,
        }


def test_beta_enable_interactive_decline_warning(root_path_populated_with_config: Path) -> None:
    root_path = root_path_populated_with_config
    with lock_and_load_config(root_path, "config.yaml") as config:
        assert "beta" not in config

    result = enable_interactive(root_path, "n\n")
    assert result.exit_code == 1
    assert result.output[-9:-1] == "Aborted!"


@pytest.mark.parametrize("write_test", [True, False])
@pytest.mark.parametrize("command", [configure, enable])
def test_beta_invalid_directories(
    root_path_populated_with_config: Path, write_test: bool, command: Callable[[Path, str, str], Result]
) -> None:
    root_path = root_path_populated_with_config
    beta_path = root_path / "beta"
    if write_test:
        (beta_path / ".write_test").mkdir(parents=True)  # `.write_test` is used in  validate_directory_writable
    if command is configure:
        generate_beta_config(root_path, True, root_path_populated_with_config)
    result = command(root_path, "--path", str(beta_path))
    assert result.exit_code == 1
    if write_test:
        assert f"Directory not writable: {str(beta_path)!r}" in result.output
    else:
        assert f"Directory doesn't exist: {str(beta_path)!r}" in result.output


@pytest.mark.parametrize("option", ["-i", "--interval"])
@pytest.mark.parametrize(
    "interval, valid",
    [
        (-1, False),
        (0, False),
        (metrics_log_interval_min - 1, False),
        (metrics_log_interval_min, True),
        (metrics_log_interval_min + 1, True),
        (metrics_log_interval_max + 1, False),
        (metrics_log_interval_max - 1, True),
        (metrics_log_interval_max, True),
    ],
)
def test_beta_configure_interval(
    root_path_populated_with_config: Path, interval: int, valid: bool, option: str
) -> None:
    root_path = root_path_populated_with_config
    beta_path = root_path / "beta"
    beta_path.mkdir()
    generate_beta_config(root_path, True, root_path_populated_with_config)
    result = configure(root_path, "--path", str(beta_path), option, str(interval))
    assert result.exit_code == 0 if valid else 1
    if not valid:
        assert f"Must be in the range of {metrics_log_interval_min}s to {metrics_log_interval_max}s." in result.output


@pytest.mark.parametrize("enabled", [True, False])
def test_beta_disable(root_path_populated_with_config: Path, enabled: bool) -> None:
    root_path = root_path_populated_with_config
    beta_path = root_path / "beta"
    generate_beta_config(root_path, enabled, beta_path)

    result = CliRunner().invoke(
        cli,
        [
            "--root-path",
            str(root_path),
            "beta",
            "disable",
        ],
    )
    if enabled:
        assert result.exit_code == 0
        assert "beta test mode disabled" in result.output
    else:
        assert result.exit_code == 1
        assert "beta test mode is not enabled" in result.output

    with lock_and_load_config(root_path, "config.yaml") as config:
        assert config["beta"] == {
            "enabled": False,
            "path": str(beta_path),
            "metrics_log_interval": metrics_log_interval_default,
        }


@pytest.mark.parametrize(
    "versions, logs, choice, exit_code, output",
    [
        (0, 0, 1, 1, "No beta logs found"),
        (1, 0, 1, 1, "No logs files found"),
        (2, 10, 3, 1, "Invalid choice: 3"),
        (2, 10, 0, 1, "Invalid choice: 0"),
        (2, 10, -1, 1, "Invalid choice: -1"),
        (4, 3, 2, 0, "Done. You can find the prepared submission data"),
    ],
)
def test_prepare_submission(
    root_path_populated_with_config: Path, versions: int, logs: int, choice: int, exit_code: int, output: str
) -> None:
    root_path = root_path_populated_with_config
    beta_path = root_path / "beta"
    beta_path.mkdir()
    generate_beta_config(root_path, True, beta_path)

    generate_example_submission_data(beta_path, versions, logs)

    result = prepare_submission(root_path, f"{choice}\n")

    assert result.exit_code == exit_code
    assert output in result.output

    if exit_code == 0:
        submission_file = next(iter(beta_path.rglob("*.zip")))
        assert submission_file.name.startswith(f"submission_{choice - 1}")
        with zipfile.ZipFile(submission_file) as zip_file:
            all_files = [Path(info.filename) for info in zip_file.filelist]
            for version in range(versions):
                chik_blockchain_logs = Path("chik-blockchain")
                plotting_logs = Path("plotting")
                for i in range(logs):
                    assert chik_blockchain_logs / f"beta_{i}.log" in all_files
                    assert chik_blockchain_logs / f"beta_{i + 10}.gz" in all_files
                    assert plotting_logs / f"plot_{i}.log" in all_files


@pytest.mark.parametrize(
    "enabled, path, interval",
    [
        (True, Path("path_1"), metrics_log_interval_min),
        (False, Path("path_2"), metrics_log_interval_max),
    ],
)
def test_beta_status(root_path_populated_with_config: Path, enabled: bool, path: Path, interval: int) -> None:
    root_path = root_path_populated_with_config
    generate_beta_config(root_path, enabled, path)

    result = CliRunner().invoke(
        cli,
        [
            "--root-path",
            str(root_path),
            "beta",
            "status",
        ],
    )

    assert result.exit_code == 0
    assert f"enabled: {enabled}" in result.output
    assert f"path: {path!s}" in result.output
    assert f"metrics log interval: {metrics_log_interval_default!s}" in result.output



from __future__ import annotations

from typing import Optional

import pytest
from chik_rs import AugSchemeMPL, G2Element, PrivateKey
from klvm.casts import int_to_bytes

from chik._tests.klvm.benchmark_costs import cost_of_spend_bundle
from chik._tests.klvm.test_puzzles import secret_exponent_for_index
from chik._tests.conftest import ConsensusMode
from chik._tests.util.spend_sim import CostLogger, SimClient, SpendSim, sim_and_client
from chik.types.blockchain_format.coin import Coin
from chik.types.blockchain_format.program import Program
from chik.types.coin_spend import make_spend
from chik.types.mempool_inclusion_status import MempoolInclusionStatus
from chik.util.errors import Err
from chik.util.ints import uint64
from chik.wallet.cat_wallet.cat_utils import (
    CAT_MOD,
    SpendableCAT,
    construct_cat_puzzle,
    unsigned_spend_bundle_for_spendable_cats,
)
from chik.wallet.lineage_proof import LineageProof
from chik.wallet.puzzles.tails import DelegatedLimitations, EverythingWithSig, GenesisById, GenesisByPuzhash
from chik.wallet.wallet_spend_bundle import WalletSpendBundle

acs = Program.to(1)
acs_ph = acs.get_tree_hash()
NO_LINEAGE_PROOF = LineageProof()


async def do_spend(
    sim: SpendSim,
    sim_client: SimClient,
    tail: Program,
    coins: list[Coin],
    lineage_proofs: list[LineageProof],
    inner_solutions: list[Program],
    expected_result: tuple[MempoolInclusionStatus, Optional[Err]],
    reveal_limitations_program: bool = True,
    signatures: list[G2Element] = [],
    extra_deltas: Optional[list[int]] = None,
    additional_spends: list[WalletSpendBundle] = [],
    limitations_solutions: Optional[list[Program]] = None,
    cost_logger: Optional[CostLogger] = None,
    cost_log_msg: str = "",
) -> int:
    if limitations_solutions is None:
        limitations_solutions = [Program.to([])] * len(coins)
    if extra_deltas is None:
        extra_deltas = [0] * len(coins)

    spendable_cat_list: list[SpendableCAT] = []
    for coin, innersol, proof, limitations_solution, extra_delta in zip(
        coins, inner_solutions, lineage_proofs, limitations_solutions, extra_deltas
    ):
        spendable_cat_list.append(
            SpendableCAT(
                coin,
                tail.get_tree_hash(),
                acs,
                innersol,
                limitations_solution=limitations_solution,
                lineage_proof=proof,
                extra_delta=extra_delta,
                limitations_program_reveal=tail if reveal_limitations_program else Program.to([]),
            )
        )

    spend_bundle = unsigned_spend_bundle_for_spendable_cats(CAT_MOD, spendable_cat_list)
    agg_sig = AugSchemeMPL.aggregate(signatures)
    final_bundle = WalletSpendBundle.aggregate(
        [*additional_spends, spend_bundle, WalletSpendBundle([], agg_sig)]  # "Signing" the spend bundle
    )
    if cost_logger is not None:
        final_bundle = cost_logger.add_cost(cost_log_msg, final_bundle)
    result = await sim_client.push_tx(final_bundle)
    assert result == expected_result
    cost = cost_of_spend_bundle(spend_bundle)
    await sim.farm_block()
    return cost


@pytest.mark.anyio
async def test_cat_mod(cost_logger: CostLogger, consensus_mode: ConsensusMode) -> None:
    async with sim_and_client() as (sim, sim_client):
        tail = Program.to([])
        checker_solution = Program.to([])
        cat_puzzle = construct_cat_puzzle(CAT_MOD, tail.get_tree_hash(), acs)
        cat_ph = cat_puzzle.get_tree_hash()
        await sim.farm_block(cat_ph)
        starting_coin = (await sim_client.get_coin_records_by_puzzle_hash(cat_ph))[0].coin

        # Testing the eve spend
        await do_spend(
            sim,
            sim_client,
            tail,
            [starting_coin],
            [NO_LINEAGE_PROOF],
            [
                Program.to(
                    [
                        [51, acs_ph, starting_coin.amount - 3, [b"memo"]],
                        [51, acs_ph, 1],
                        [51, acs_ph, 2],
                        [51, 0, -113, tail, checker_solution],
                    ]
                )
            ],
            (MempoolInclusionStatus.SUCCESS, None),
            limitations_solutions=[checker_solution],
            cost_logger=cost_logger,
            cost_log_msg="Cat Eve Spend + create three children (TAIL: ())",
        )

        # There's 4 total coins at this point. A farming reward and the three children of the spend above.

        # Testing a combination of two
        coins = [
            record.coin
            for record in (await sim_client.get_coin_records_by_puzzle_hash(cat_ph, include_spent_coins=False))
        ]
        coins = [coins[0], coins[1]]
        await do_spend(
            sim,
            sim_client,
            tail,
            coins,
            [NO_LINEAGE_PROOF] * 2,
            [
                Program.to([[51, acs_ph, coins[0].amount + coins[1].amount], [51, 0, -113, tail, checker_solution]]),
                Program.to([[51, 0, -113, tail, checker_solution]]),
            ],
            (MempoolInclusionStatus.SUCCESS, None),
            limitations_solutions=[checker_solution] * 2,
            cost_logger=cost_logger,
            cost_log_msg="Cat Spend x2 + create one child (TAIL: ())",
        )

        # Testing a combination of three
        coins = [
            record.coin
            for record in (await sim_client.get_coin_records_by_puzzle_hash(cat_ph, include_spent_coins=False))
        ]
        total_amount = uint64(sum(c.amount for c in coins))
        await do_spend(
            sim,
            sim_client,
            tail,
            coins,
            [NO_LINEAGE_PROOF] * 3,
            [
                Program.to([[51, acs_ph, total_amount], [51, 0, -113, tail, checker_solution]]),
                Program.to([[51, 0, -113, tail, checker_solution]]),
                Program.to([[51, 0, -113, tail, checker_solution]]),
            ],
            (MempoolInclusionStatus.SUCCESS, None),
            limitations_solutions=[checker_solution] * 3,
            cost_logger=cost_logger,
            cost_log_msg="Cat Spend x3 + create one child (TAIL: ())",
        )

        # Spend with a standard lineage proof
        parent_coin = coins[0]  # The first one is the one we didn't light on fire
        _, curried_args = cat_puzzle.uncurry()
        _, _, innerpuzzle = curried_args.as_iter()
        lineage_proof = LineageProof(
            parent_coin.parent_coin_info, innerpuzzle.get_tree_hash(), uint64(parent_coin.amount)
        )
        await do_spend(
            sim,
            sim_client,
            tail,
            [(await sim_client.get_coin_records_by_puzzle_hash(cat_ph, include_spent_coins=False))[0].coin],
            [lineage_proof],
            [Program.to([[51, acs_ph, total_amount]])],
            (MempoolInclusionStatus.SUCCESS, None),
            reveal_limitations_program=False,
            cost_logger=cost_logger,
            cost_log_msg="Cat Spend + create one child",
        )

        # Melt some value
        await do_spend(
            sim,
            sim_client,
            tail,
            [(await sim_client.get_coin_records_by_puzzle_hash(cat_ph, include_spent_coins=False))[0].coin],
            [NO_LINEAGE_PROOF],
            [Program.to([[51, acs_ph, total_amount - 1], [51, 0, -113, tail, checker_solution]])],
            (MempoolInclusionStatus.SUCCESS, None),
            extra_deltas=[-1],
            limitations_solutions=[checker_solution],
            cost_logger=cost_logger,
            cost_log_msg="Cat Spend (Melt) + create one child (TAIL: ())",
        )

        # Mint some value
        await sim.farm_block(acs_ph)
        acs_coin = (await sim_client.get_coin_records_by_puzzle_hash(acs_ph, include_spent_coins=False))[0].coin
        acs_bundle = WalletSpendBundle([make_spend(acs_coin, acs, Program.to([]))], G2Element())
        await do_spend(
            sim,
            sim_client,
            tail,
            [(await sim_client.get_coin_records_by_puzzle_hash(cat_ph, include_spent_coins=False))[0].coin],
            [NO_LINEAGE_PROOF],
            [
                Program.to([[51, acs_ph, total_amount], [51, 0, -113, tail, checker_solution]])
            ],  # We subtracted 1 last time so it's normal now
            (MempoolInclusionStatus.SUCCESS, None),
            extra_deltas=[1],
            additional_spends=[acs_bundle],
            limitations_solutions=[checker_solution],
            cost_logger=cost_logger,
            cost_log_msg="ACS burn + Cat Spend (Mint) + create one child (TAIL: ())",
        )


@pytest.mark.anyio
async def test_complex_spend(cost_logger: CostLogger, consensus_mode: ConsensusMode) -> None:
    async with sim_and_client() as (sim, sim_client):
        tail = Program.to([])
        checker_solution = Program.to([])
        cat_puzzle = construct_cat_puzzle(CAT_MOD, tail.get_tree_hash(), acs)
        cat_ph = cat_puzzle.get_tree_hash()
        await sim.farm_block(cat_ph)
        await sim.farm_block(cat_ph)

        cat_records = await sim_client.get_coin_records_by_puzzle_hash(cat_ph, include_spent_coins=False)
        parent_of_mint = cat_records[0].coin
        parent_of_melt = cat_records[1].coin
        eve_to_mint = cat_records[2].coin
        eve_to_melt = cat_records[3].coin

        # Spend two of them to make them non-eve
        await do_spend(
            sim,
            sim_client,
            tail,
            [parent_of_mint, parent_of_melt],
            [NO_LINEAGE_PROOF, NO_LINEAGE_PROOF],
            [
                Program.to([[51, acs_ph, parent_of_mint.amount], [51, 0, -113, tail, checker_solution]]),
                Program.to([[51, acs_ph, parent_of_melt.amount], [51, 0, -113, tail, checker_solution]]),
            ],
            (MempoolInclusionStatus.SUCCESS, None),
            limitations_solutions=[checker_solution] * 2,
            cost_logger=cost_logger,
            cost_log_msg="Cat Eve Spend x2 + create one child each (TAIL: ())",
        )

        # Make the lineage proofs for the non-eves
        mint_lineage = LineageProof(parent_of_mint.parent_coin_info, acs_ph, uint64(parent_of_mint.amount))
        melt_lineage = LineageProof(parent_of_melt.parent_coin_info, acs_ph, uint64(parent_of_melt.amount))

        # Find the two new coins
        all_cats = await sim_client.get_coin_records_by_puzzle_hash(cat_ph, include_spent_coins=False)
        all_cat_coins = [cr.coin for cr in all_cats]
        standard_to_mint = next(filter(lambda cr: cr.parent_coin_info == parent_of_mint.name(), all_cat_coins))
        standard_to_melt = next(filter(lambda cr: cr.parent_coin_info == parent_of_melt.name(), all_cat_coins))

        # Do the complex spend
        # We have both and eve and non-eve doing both minting and melting
        await do_spend(
            sim,
            sim_client,
            tail,
            [eve_to_mint, eve_to_melt, standard_to_mint, standard_to_melt],
            [NO_LINEAGE_PROOF, NO_LINEAGE_PROOF, mint_lineage, melt_lineage],
            [
                Program.to([[51, acs_ph, eve_to_mint.amount + 13], [51, 0, -113, tail, checker_solution]]),
                Program.to([[51, acs_ph, eve_to_melt.amount - 21], [51, 0, -113, tail, checker_solution]]),
                Program.to([[51, acs_ph, standard_to_mint.amount + 21], [51, 0, -113, tail, checker_solution]]),
                Program.to([[51, acs_ph, standard_to_melt.amount - 13], [51, 0, -113, tail, checker_solution]]),
            ],
            (MempoolInclusionStatus.SUCCESS, None),
            limitations_solutions=[checker_solution] * 4,
            extra_deltas=[13, -21, 21, -13],
            cost_logger=cost_logger,
            cost_log_msg="Cat Eve Spend x2 (mint & melt) + Cat Spend x2 (mint & melt) - one child each (TAIL: ())",
        )


@pytest.mark.anyio
async def test_genesis_by_id(cost_logger: CostLogger, consensus_mode: ConsensusMode) -> None:
    async with sim_and_client() as (sim, sim_client):
        await sim.farm_block(acs_ph)

        starting_coin = (await sim_client.get_coin_records_by_puzzle_hash(acs_ph))[0].coin
        tail = GenesisById.construct([Program.to(starting_coin.name())])
        checker_solution = GenesisById.solve([], {})
        cat_puzzle = construct_cat_puzzle(CAT_MOD, tail.get_tree_hash(), acs)
        cat_ph = cat_puzzle.get_tree_hash()

        await sim_client.push_tx(
            WalletSpendBundle(
                [make_spend(starting_coin, acs, Program.to([[51, cat_ph, starting_coin.amount]]))], G2Element()
            )
        )
        await sim.farm_block()

        await do_spend(
            sim,
            sim_client,
            tail,
            [(await sim_client.get_coin_records_by_puzzle_hash(cat_ph, include_spent_coins=False))[0].coin],
            [NO_LINEAGE_PROOF],
            [Program.to([[51, acs_ph, starting_coin.amount], [51, 0, -113, tail, checker_solution]])],
            (MempoolInclusionStatus.SUCCESS, None),
            limitations_solutions=[checker_solution],
            cost_logger=cost_logger,
            cost_log_msg="Cat Eve Spend - create one child (TAIL: genesis_by_id)",
        )


@pytest.mark.anyio
async def test_genesis_by_puzhash(cost_logger: CostLogger, consensus_mode: ConsensusMode) -> None:
    async with sim_and_client() as (sim, sim_client):
        await sim.farm_block(acs_ph)

        starting_coin = (await sim_client.get_coin_records_by_puzzle_hash(acs_ph))[0].coin
        tail = GenesisByPuzhash.construct([Program.to(starting_coin.puzzle_hash)])
        checker_solution = GenesisByPuzhash.solve([], starting_coin.to_json_dict())
        cat_puzzle = construct_cat_puzzle(CAT_MOD, tail.get_tree_hash(), acs)
        cat_ph = cat_puzzle.get_tree_hash()

        await sim_client.push_tx(
            WalletSpendBundle(
                [make_spend(starting_coin, acs, Program.to([[51, cat_ph, starting_coin.amount]]))], G2Element()
            )
        )
        await sim.farm_block()

        await do_spend(
            sim,
            sim_client,
            tail,
            [(await sim_client.get_coin_records_by_puzzle_hash(cat_ph, include_spent_coins=False))[0].coin],
            [NO_LINEAGE_PROOF],
            [Program.to([[51, acs_ph, starting_coin.amount], [51, 0, -113, tail, checker_solution]])],
            (MempoolInclusionStatus.SUCCESS, None),
            limitations_solutions=[checker_solution],
            cost_logger=cost_logger,
            cost_log_msg="Cat Eve Spend - create one child (TAIL: genesis_by_puzhash)",
        )


@pytest.mark.anyio
async def test_everything_with_signature(cost_logger: CostLogger, consensus_mode: ConsensusMode) -> None:
    async with sim_and_client() as (sim, sim_client):
        sk = PrivateKey.from_bytes(secret_exponent_for_index(1).to_bytes(32, "big"))
        tail = EverythingWithSig.construct([Program.to(sk.get_g1())])
        checker_solution = EverythingWithSig.solve([], {})
        cat_puzzle = construct_cat_puzzle(CAT_MOD, tail.get_tree_hash(), acs)
        cat_ph = cat_puzzle.get_tree_hash()
        await sim.farm_block(cat_ph)

        # Test eve spend
        # We don't sign any message data because KLVM 0 translates to b'' apparently
        starting_coin = (await sim_client.get_coin_records_by_puzzle_hash(cat_ph))[0].coin
        signature = AugSchemeMPL.sign(sk, (starting_coin.name() + sim.defaults.AGG_SIG_ME_ADDITIONAL_DATA))

        await do_spend(
            sim,
            sim_client,
            tail,
            [starting_coin],
            [NO_LINEAGE_PROOF],
            [Program.to([[51, acs_ph, starting_coin.amount], [51, 0, -113, tail, checker_solution]])],
            (MempoolInclusionStatus.SUCCESS, None),
            limitations_solutions=[checker_solution],
            signatures=[signature],
            cost_logger=cost_logger,
            cost_log_msg="Cat Eve Spend - create one child (TAIL: everything_with_signature)",
        )

        # Test melting value
        coin = (await sim_client.get_coin_records_by_puzzle_hash(cat_ph, include_spent_coins=False))[0].coin
        signature = AugSchemeMPL.sign(sk, (int_to_bytes(-1) + coin.name() + sim.defaults.AGG_SIG_ME_ADDITIONAL_DATA))

        await do_spend(
            sim,
            sim_client,
            tail,
            [coin],
            [NO_LINEAGE_PROOF],
            [Program.to([[51, acs_ph, coin.amount - 1], [51, 0, -113, tail, checker_solution]])],
            (MempoolInclusionStatus.SUCCESS, None),
            extra_deltas=[-1],
            limitations_solutions=[checker_solution],
            signatures=[signature],
            cost_logger=cost_logger,
            cost_log_msg="Cat Spend (Melt) - create one child (TAIL: everything_with_signature)",
        )

        # Test minting value
        coin = (await sim_client.get_coin_records_by_puzzle_hash(cat_ph, include_spent_coins=False))[0].coin
        signature = AugSchemeMPL.sign(sk, (int_to_bytes(1) + coin.name() + sim.defaults.AGG_SIG_ME_ADDITIONAL_DATA))

        # Need something to fund the minting
        await sim.farm_block(acs_ph)
        acs_coin = (await sim_client.get_coin_records_by_puzzle_hash(acs_ph, include_spent_coins=False))[0].coin
        acs_bundle = WalletSpendBundle([make_spend(acs_coin, acs, Program.to([]))], G2Element())

        await do_spend(
            sim,
            sim_client,
            tail,
            [coin],
            [NO_LINEAGE_PROOF],
            [Program.to([[51, acs_ph, coin.amount + 1], [51, 0, -113, tail, checker_solution]])],
            (MempoolInclusionStatus.SUCCESS, None),
            extra_deltas=[1],
            limitations_solutions=[checker_solution],
            signatures=[signature],
            additional_spends=[acs_bundle],
            cost_logger=cost_logger,
            cost_log_msg="ACS Burn + Cat Spend (Mint) - create one child (TAIL: everything_with_signature)",
        )


@pytest.mark.anyio
async def test_delegated_tail(cost_logger: CostLogger, consensus_mode: ConsensusMode) -> None:
    async with sim_and_client() as (sim, sim_client):
        await sim.farm_block(acs_ph)

        starting_coin = (await sim_client.get_coin_records_by_puzzle_hash(acs_ph))[0].coin
        sk = PrivateKey.from_bytes(secret_exponent_for_index(1).to_bytes(32, "big"))
        tail = DelegatedLimitations.construct([Program.to(sk.get_g1())])
        cat_puzzle = construct_cat_puzzle(CAT_MOD, tail.get_tree_hash(), acs)
        cat_ph = cat_puzzle.get_tree_hash()

        await sim_client.push_tx(
            WalletSpendBundle(
                [make_spend(starting_coin, acs, Program.to([[51, cat_ph, starting_coin.amount]]))], G2Element()
            )
        )
        await sim.farm_block()

        # We're signing a different tail to use here
        name_as_program = Program.to(starting_coin.name())
        new_tail = GenesisById.construct([name_as_program])
        checker_solution = DelegatedLimitations.solve(
            [name_as_program],
            {
                "signed_program": {"identifier": "genesis_by_id", "args": [str(name_as_program)]},
                "program_arguments": {},
            },
        )
        signature = AugSchemeMPL.sign(sk, new_tail.get_tree_hash())

        await do_spend(
            sim,
            sim_client,
            tail,
            [(await sim_client.get_coin_records_by_puzzle_hash(cat_ph, include_spent_coins=False))[0].coin],
            [NO_LINEAGE_PROOF],
            [Program.to([[51, acs_ph, starting_coin.amount], [51, 0, -113, tail, checker_solution]])],
            (MempoolInclusionStatus.SUCCESS, None),
            signatures=[signature],
            limitations_solutions=[checker_solution],
            cost_logger=cost_logger,
            cost_log_msg="Cat Eve Spend - create one child (TAIL: delegated_tail - genesis_by_id)",
        )



from __future__ import annotations

import pytest

from chik.rpc.crawler_rpc_api import CrawlerRpcApi
from chik.seeder.crawler import Crawler


@pytest.mark.anyio
async def test_get_ips_after_timestamp(bt):
    crawler = Crawler(bt.config.get("seeder", {}), bt.root_path, constants=bt.constants)
    crawler_rpc_api = CrawlerRpcApi(crawler)

    # Should raise ValueError when `after` is not supplied
    with pytest.raises(ValueError):
        await crawler_rpc_api.get_ips_after_timestamp({})

    # Crawler isn't actually crawling, so this should return zero IPs
    response = await crawler_rpc_api.get_ips_after_timestamp({"after": 0})
    assert len(response["ips"]) == 0

    # Add some known data
    # IPs are listed here out of order (by time) to test consistent sorting
    # Timestamps increase as the IP value increases
    crawler.best_timestamp_per_peer["0.0.0.0"] = 0
    crawler.best_timestamp_per_peer["2.2.2.2"] = 1644300000
    crawler.best_timestamp_per_peer["1.1.1.1"] = 1644213600
    crawler.best_timestamp_per_peer["7.7.7.7"] = 1644732000
    crawler.best_timestamp_per_peer["3.3.3.3"] = 1644386400
    crawler.best_timestamp_per_peer["4.4.4.4"] = 1644472800
    crawler.best_timestamp_per_peer["9.9.9.9"] = 1644904800
    crawler.best_timestamp_per_peer["5.5.5.5"] = 1644559200
    crawler.best_timestamp_per_peer["6.6.6.6"] = 1644645600
    crawler.best_timestamp_per_peer["8.8.8.8"] = 1644818400

    response = await crawler_rpc_api.get_ips_after_timestamp({"after": 0})
    assert len(response["ips"]) == 9

    response = await crawler_rpc_api.get_ips_after_timestamp({"after": 1644473000})
    assert len(response["ips"]) == 5

    # Test offset/limit functionality
    response = await crawler_rpc_api.get_ips_after_timestamp({"after": 0, "limit": 2})
    assert len(response["ips"]) == 2
    assert response["total"] == 9
    assert response["ips"][0] == "1.1.1.1"
    assert response["ips"][1] == "2.2.2.2"

    response = await crawler_rpc_api.get_ips_after_timestamp({"after": 0, "offset": 2, "limit": 2})
    assert len(response["ips"]) == 2
    assert response["total"] == 9
    assert response["ips"][0] == "3.3.3.3"
    assert response["ips"][1] == "4.4.4.4"



from __future__ import annotations

from typing import Any, Union

import pytest

from chik.types.blockchain_format.program import Program
from chik.wallet.puzzle_drivers import PuzzleInfo, Solver


def test_puzzle_info() -> None:
    test_driver: dict[str, Any] = {
        "string": "hello",
        "bytes": "0xcafef00d",
        "int": "123",
        "program": "(q . 'hello')",
        "zero": "0",
        "nil": "()",
    }
    test_also: dict[str, Any] = {"type": "TEST", "string": "hello"}
    test_driver["also"] = test_also

    with pytest.raises(ValueError, match="A type is required"):
        PuzzleInfo(test_driver)
    solver = Solver(test_driver)

    test_driver["type"] = "TEST"
    puzzle_info = PuzzleInfo(test_driver)

    assert puzzle_info.type() == "TEST"
    assert puzzle_info.also() == PuzzleInfo(test_also)

    capitalize_bytes = test_driver.copy()
    capitalize_bytes["bytes"] = "0xCAFEF00D"
    assert solver == Solver(capitalize_bytes)
    assert puzzle_info == PuzzleInfo(capitalize_bytes)

    obj: Union[PuzzleInfo, Solver]
    for obj in (puzzle_info, solver):
        assert obj["string"] == "hello"
        assert obj["bytes"] == bytes.fromhex("cafef00d")
        assert obj["int"] == 123
        assert obj["program"] == Program.to((1, "hello"))
        assert obj["zero"] == 0
        assert obj["nil"] == Program.to([])



from __future__ import annotations

import pytest
from chik_rs import G2Element

from chik._tests.util.spend_sim import CostLogger, sim_and_client
from chik.types.blockchain_format.coin import Coin
from chik.types.blockchain_format.program import Program
from chik.types.blockchain_format.sized_bytes import bytes32
from chik.types.coin_spend import make_spend
from chik.types.mempool_inclusion_status import MempoolInclusionStatus
from chik.util.errors import Err
from chik.wallet.puzzles.load_klvm import load_klvm
from chik.wallet.util.merkle_utils import build_merkle_tree, build_merkle_tree_from_binary_tree, simplify_merkle_proof
from chik.wallet.wallet_spend_bundle import WalletSpendBundle

GRAFTROOT_MOD = load_klvm("graftroot_dl_offers.clsp", package_or_requirement="chik.data_layer.puzzles")

# Always returns the last value
# (mod solution
#
#   (defun recurse (solution last_value)
#     (if solution
#         (recurse (r solution) (f solution))
#         last_value
#     )
#   )
#
#   (recurse solution ())
# )
ACS = Program.fromhex(
    "ff02ffff01ff02ff02ffff04ff02ffff04ff03ffff01ff8080808080ffff04ffff01ff02ffff03ff05ffff01ff02ff02ffff04ff02ffff04ff0dffff04ff09ff8080808080ffff010b80ff0180ff018080"
)
ACS_PH = ACS.get_tree_hash()

NIL_PH = Program.to(None).get_tree_hash()


@pytest.mark.anyio
async def test_graftroot(cost_logger: CostLogger) -> None:
    async with sim_and_client() as (sim, sim_client):
        # Create the coin we're testing
        all_values: list[bytes32] = [bytes32([x] * 32) for x in range(0, 100)]
        root, proofs = build_merkle_tree(all_values)
        p2_conditions = Program.to((1, [[51, ACS_PH, 0]]))  # An coin to create to make sure this hits the blockchain
        desired_key_values = ((bytes32.zeros, bytes32([1] * 32)), (bytes32([7] * 32), bytes32([8] * 32)))
        desired_row_hashes: list[bytes32] = [build_merkle_tree_from_binary_tree(kv)[0] for kv in desired_key_values]
        fake_struct: Program = Program.to((ACS_PH, NIL_PH))
        graftroot_puzzle: Program = GRAFTROOT_MOD.curry(
            # Do everything twice to test depending on multiple singleton updates
            p2_conditions,
            [fake_struct, fake_struct],
            [ACS_PH, ACS_PH],
            [desired_row_hashes, desired_row_hashes],
        )
        await sim.farm_block(graftroot_puzzle.get_tree_hash())
        graftroot_coin: Coin = (await sim_client.get_coin_records_by_puzzle_hash(graftroot_puzzle.get_tree_hash()))[
            0
        ].coin

        # Build some merkle trees that won't satidy the requirements
        def filter_all(values: list[bytes32]) -> list[bytes32]:
            return [h for i, h in enumerate(values) if (h, values[min(i, i + 1)]) not in desired_key_values]

        def filter_to_only_one(values: list[bytes32]) -> list[bytes32]:
            return [h for i, h in enumerate(values) if (h, values[min(i, i + 1)]) not in desired_key_values[1:]]

        # And one that will
        def filter_none(values: list[bytes32]) -> list[bytes32]:
            return values

        for list_filter in (filter_all, filter_to_only_one, filter_none):
            # Create the "singleton"
            filtered_values = list_filter(all_values)
            root, proofs = build_merkle_tree(filtered_values)
            filtered_row_hashes: dict[bytes32, tuple[int, list[bytes32]]] = {
                simplify_merkle_proof(v, (proofs[v][0], [proofs[v][1][0]])): (proofs[v][0] >> 1, proofs[v][1][1:])
                for v in filtered_values
            }
            fake_puzzle: Program = ACS.curry(fake_struct, ACS.curry(ACS_PH, (root, None), NIL_PH, None))
            await sim.farm_block(fake_puzzle.get_tree_hash())
            fake_coin: Coin = (await sim_client.get_coin_records_by_puzzle_hash(fake_puzzle.get_tree_hash()))[0].coin

            # Create the spend
            fake_spend = make_spend(
                fake_coin,
                fake_puzzle,
                Program.to([[[62, "$"]]]),
            )

            proofs_of_inclusion = []
            for row_hash in desired_row_hashes:
                if row_hash in filtered_row_hashes:
                    proofs_of_inclusion.append(filtered_row_hashes[row_hash])
                else:
                    proofs_of_inclusion.append((0, []))

            graftroot_spend = make_spend(
                graftroot_coin,
                graftroot_puzzle,
                Program.to(
                    [
                        # Again, everything twice
                        [proofs_of_inclusion] * 2,
                        [(root, None), (root, None)],
                        [NIL_PH, NIL_PH],
                        [NIL_PH, NIL_PH],
                        [],
                    ]
                ),
            )

            final_bundle = WalletSpendBundle([fake_spend, graftroot_spend], G2Element())
            result = await sim_client.push_tx(final_bundle)

            # If this is the satisfactory merkle tree
            if filtered_values == all_values:
                cost_logger.add_cost(
                    "DL Graftroot - fake singleton w/ announce + prove two rows in a DL merkle tree + create one child",
                    final_bundle,
                )
                assert result == (MempoolInclusionStatus.SUCCESS, None)
                # clear the mempool
                same_height = sim.block_height
                await sim.farm_block()
                assert len(await sim_client.get_coin_records_by_puzzle_hash(ACS_PH)) > 0
                await sim.rewind(same_height)

                # try with a bad merkle root announcement
                new_fake_spend = make_spend(
                    fake_coin,
                    ACS.curry(fake_struct, ACS.curry(ACS_PH, (bytes32.zeros, None), None, None)),
                    Program.to([[[62, "$"]]]),
                )
                new_final_bundle = WalletSpendBundle([new_fake_spend, graftroot_spend], G2Element())
                result = await sim_client.push_tx(new_final_bundle)
                assert result == (MempoolInclusionStatus.FAILED, Err.ASSERT_ANNOUNCE_CONSUMED_FAILED)
            else:
                assert result == (MempoolInclusionStatus.FAILED, Err.GENERATOR_RUNTIME_ERROR)
                with pytest.raises(ValueError, match="klvm raise"):
                    graftroot_puzzle.run(graftroot_spend.solution.to_program())



from __future__ import annotations

from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Optional, cast
from unittest.mock import MagicMock

import pytest
from chik_rs import G1Element

from chik._tests.util.benchmarks import rand_g1, rand_hash
from chik.pools.pool_wallet import PoolWallet
from chik.types.blockchain_format.sized_bytes import bytes32


@dataclass
class MockStandardWallet:
    canned_puzzlehash: bytes32

    async def get_new_puzzlehash(self) -> bytes32:
        return self.canned_puzzlehash

    async def get_puzzle_hash(self, new: bool) -> bytes32:
        return self.canned_puzzlehash


@dataclass
class MockWalletStateManager:
    root_path: Optional[Path] = None
    config: dict[str, Any] = field(default_factory=dict)


@dataclass
class MockPoolWalletConfig:
    launcher_id: bytes32
    pool_url: str
    payout_instructions: str
    target_puzzle_hash: bytes32
    p2_singleton_puzzle_hash: bytes32
    owner_public_key: G1Element


@dataclass
class MockPoolState:
    pool_url: Optional[str]
    target_puzzle_hash: bytes32
    owner_pubkey: G1Element


@dataclass
class MockPoolWalletInfo:
    launcher_id: bytes32
    p2_singleton_puzzle_hash: bytes32
    current: MockPoolState


@pytest.mark.anyio
async def test_update_pool_config_new_config(monkeypatch: Any) -> None:
    """
    Test that PoolWallet can create a new pool config
    """

    updated_configs: list[MockPoolWalletConfig] = []
    payout_instructions_ph = rand_hash()
    launcher_id: bytes32 = rand_hash()
    p2_singleton_puzzle_hash: bytes32 = rand_hash()
    pool_url: str = ""
    target_puzzle_hash: bytes32 = rand_hash()
    owner_pubkey: G1Element = rand_g1()
    current: MockPoolState = MockPoolState(
        pool_url=pool_url,
        target_puzzle_hash=target_puzzle_hash,
        owner_pubkey=owner_pubkey,
    )
    current_state: MockPoolWalletInfo = MockPoolWalletInfo(
        launcher_id=launcher_id,
        p2_singleton_puzzle_hash=p2_singleton_puzzle_hash,
        current=current,
    )

    # No config data
    def mock_load_pool_config(root_path: Path) -> list[MockPoolWalletConfig]:
        return []

    monkeypatch.setattr("chik.pools.pool_wallet.load_pool_config", mock_load_pool_config)

    # Mock pool_config.update_pool_config to capture the updated configs
    async def mock_pool_config_update_pool_config(
        root_path: Path, pool_config_list: list[MockPoolWalletConfig]
    ) -> None:
        nonlocal updated_configs
        updated_configs = pool_config_list

    monkeypatch.setattr("chik.pools.pool_wallet.update_pool_config", mock_pool_config_update_pool_config)

    # Mock PoolWallet.get_current_state to return our canned state
    async def mock_get_current_state(self: Any) -> Any:
        return current_state

    monkeypatch.setattr(PoolWallet, "get_current_state", mock_get_current_state)

    # Create an empty PoolWallet and populate only the required fields
    wallet = PoolWallet(
        wallet_state_manager=MockWalletStateManager(),  # type: ignore[arg-type]
        standard_wallet=cast(Any, MockStandardWallet(canned_puzzlehash=payout_instructions_ph)),
        log=MagicMock(),
        wallet_info=MagicMock(),
        wallet_id=MagicMock(),
    )

    await wallet.update_pool_config()

    assert len(updated_configs) == 1
    assert updated_configs[0].launcher_id == launcher_id
    assert updated_configs[0].pool_url == pool_url
    assert updated_configs[0].payout_instructions == payout_instructions_ph.hex()
    assert updated_configs[0].target_puzzle_hash == target_puzzle_hash
    assert updated_configs[0].p2_singleton_puzzle_hash == p2_singleton_puzzle_hash
    assert updated_configs[0].owner_public_key == owner_pubkey


@pytest.mark.anyio
async def test_update_pool_config_existing_payout_instructions(monkeypatch: Any) -> None:
    """
    Test that PoolWallet will retain existing payout_instructions when updating the pool config.
    """

    updated_configs: list[MockPoolWalletConfig] = []
    payout_instructions_ph = rand_hash()
    launcher_id: bytes32 = rand_hash()
    p2_singleton_puzzle_hash: bytes32 = rand_hash()
    pool_url: str = "https://fake.pool.url"
    target_puzzle_hash: bytes32 = rand_hash()
    owner_pubkey: G1Element = rand_g1()
    current: MockPoolState = MockPoolState(
        pool_url=pool_url,
        target_puzzle_hash=target_puzzle_hash,
        owner_pubkey=owner_pubkey,
    )
    current_state: MockPoolWalletInfo = MockPoolWalletInfo(
        launcher_id=launcher_id,
        p2_singleton_puzzle_hash=p2_singleton_puzzle_hash,
        current=current,
    )

    # Existing config data with different values
    # payout_instructions should _NOT_ be updated after calling update_pool_config
    existing_launcher_id: bytes32 = launcher_id
    existing_pool_url: str = ""
    existing_payout_instructions_ph: bytes32 = rand_hash()
    existing_target_puzzle_hash: bytes32 = rand_hash()
    existing_p2_singleton_puzzle_hash: bytes32 = rand_hash()
    existing_owner_pubkey: G1Element = rand_g1()
    existing_config: MockPoolWalletConfig = MockPoolWalletConfig(
        launcher_id=existing_launcher_id,
        pool_url=existing_pool_url,
        payout_instructions=existing_payout_instructions_ph.hex(),
        target_puzzle_hash=existing_target_puzzle_hash,
        p2_singleton_puzzle_hash=existing_p2_singleton_puzzle_hash,
        owner_public_key=existing_owner_pubkey,
    )

    # No config data
    def mock_load_pool_config(root_path: Path) -> list[MockPoolWalletConfig]:
        nonlocal existing_config
        return [existing_config]

    monkeypatch.setattr("chik.pools.pool_wallet.load_pool_config", mock_load_pool_config)

    # Mock pool_config.update_pool_config to capture the updated configs
    async def mock_pool_config_update_pool_config(
        root_path: Path, pool_config_list: list[MockPoolWalletConfig]
    ) -> None:
        nonlocal updated_configs
        updated_configs = pool_config_list

    monkeypatch.setattr("chik.pools.pool_wallet.update_pool_config", mock_pool_config_update_pool_config)

    # Mock PoolWallet.get_current_state to return our canned state
    async def mock_get_current_state(self: Any) -> Any:
        return current_state

    monkeypatch.setattr(PoolWallet, "get_current_state", mock_get_current_state)

    # Create an empty PoolWallet and populate only the required fields
    wallet = PoolWallet(
        wallet_state_manager=MockWalletStateManager(),  # type: ignore[arg-type]
        standard_wallet=cast(Any, MockStandardWallet(canned_puzzlehash=payout_instructions_ph)),
        log=MagicMock(),
        wallet_info=MagicMock(),
        wallet_id=MagicMock(),
    )

    await wallet.update_pool_config()

    assert len(updated_configs) == 1
    assert updated_configs[0].launcher_id == launcher_id
    assert updated_configs[0].pool_url == pool_url

    # payout_instructions should still point to existing_payout_instructions_ph
    assert updated_configs[0].payout_instructions == existing_payout_instructions_ph.hex()

    assert updated_configs[0].target_puzzle_hash == target_puzzle_hash
    assert updated_configs[0].p2_singleton_puzzle_hash == p2_singleton_puzzle_hash
    assert updated_configs[0].owner_public_key == owner_pubkey



from __future__ import annotations

import json
import pathlib
import subprocess
import sys

import chik._tests

build_job_matrix_path = pathlib.Path(chik._tests.__file__).with_name("build-job-matrix.py")


def run(args: list[str]) -> str:
    completed_process = subprocess.run(
        [sys.executable, build_job_matrix_path, *args],
        check=True,
        encoding="utf-8",
        stdout=subprocess.PIPE,
    )
    return completed_process.stdout


def test() -> None:
    timeouts: dict[int, dict[str, int]] = {}

    multipliers = [1, 2, 3]

    for multiplier in multipliers:
        timeouts[multiplier] = {}
        output = run(args=["--per", "directory", "--timeout-multiplier", str(multiplier)])
        matrix = json.loads(output)
        for entry in matrix:
            timeouts[multiplier][entry["name"]] = entry["job_timeout"]

    reference = timeouts[1]

    for multiplier in multipliers:
        if multiplier == 1:
            continue

        adjusted_reference = {key: value * multiplier for key, value in reference.items()}
        assert timeouts[multiplier] == adjusted_reference
